@@ -49,15 +49,40 @@ void leon_init(ambapp_core *core, int level)
 }
 
 static int irqmp_initied = 0;
+static int eirqs[16];
 
-void irqmp_init(ambapp_core *core, int level)
+int irqmp_dev(ambapp_dev *dev)
 {
-	/* Only process the first GPTIMER core */
-	if (irqmp_initied)
-		return;
+	int i, c, icsel, eirq;
+	ambapp_apbdev *apbslv;
+	struct irqmp_regs *regs, *cregs;
 
-	irqmp_initied = 1;
-	irqmp = (struct irqmp_regs *)core->apbslv->address;
+	apbslv = (ambapp_apbdev *)dev;
+	regs = (struct irqmp_regs *)apbslv->address;
+	if (!irqmp_initied) {
+		/* Only process the first IRQMP core */
+		irqmp_initied = 1;
+		irqmp = regs;
+	}
+
+	/* Record eirqs, so that grgpio can avoid collisions */
+	for (c = 0; c < 16; c++) {
+		/* Logic lifted from leon_kernel.c in linux kernel */
+		icsel = BYPASS_LOAD_PA(&regs->icsel[c/8]);
+		icsel = (icsel >> ((7 - (c & 0x7)) * 4)) & 0xf;
+		cregs = regs + icsel;
+
+		eirq = (BYPASS_LOAD_PA(&cregs->mpstatus) >> 16) & 0xf;
+		for (i=0; eirq && i < 32; i++) {
+			if (!eirqs[i]) {
+				eirqs[i] = eirq;
+				break;
+			} else if (eirqs[i] == eirq) {
+				break;
+			}
+		}
+	}
+	return 0;
 }
 
 static int gptimer_inited = 0;
@@ -85,6 +110,346 @@ int gptimer_apb_fixup(ambapp_dev *dev)
 	return 0;
 }
 
+struct mutable_ambapp_pnp_ahb {
+	unsigned int	id;		/* VENDOR, DEVICE, VER, IRQ, */
+	unsigned int	custom[3];
+	unsigned int	mbar[4];	/* MASK, ADDRESS, TYPE,
+					 * CACHABLE/PREFETCHABLE */
+};
+
+unsigned int canoc_ahb_fixup_dev_count(struct ambapp_pnp_axb *axb, void *bus, int type)
+{
+	int cores = ambapp_pnp_ver(axb->id) + 1;
+	return cores - 1 ;
+}
+
+int canoc_ahb_fixup_virtual(ambapp_dev *dev, void *axb, void *bus, void **pdevp_)
+{
+	ambapp_ahbdev **pdevp = (ambapp_ahbdev **)pdevp_;
+	int i;
+	struct ambapp_pnp_ahb *orig_ahb = (struct ambapp_pnp_ahb *)axb;
+	struct mutable_ambapp_pnp_ahb ahb_buf;
+	int cores = dev->ver+1;
+	dev = NULL; /* We will overwrite it in a moment */
+
+	_memcpy((char *)&ahb_buf, (char *)orig_ahb, sizeof(struct ambapp_pnp_ahb));
+	ahb_buf.id &= ~0x3e0;		/* Set version to 0 */
+	ahb_buf.mbar[0] |= 0xfff0;	/* Set MASK to 0xfff */
+
+	for (i = 0; i < cores; i++) {
+		if(i > 0) {
+			/* Overwrites actual device entry at i == 0 */
+			(*pdevp)++;
+		}
+		(*pdevp)->dev_type = DEV_AHB_SLV;
+		ambapp_dev_init_ahb((struct ahb_bus*)bus, (struct ambapp_pnp_ahb *)&ahb_buf, *pdevp);
+		ahb_buf.id++;			/* Increase irq */
+		ahb_buf.mbar[0] += 0x100000;	/* Increase address */
+	}
+
+	return 1;
+}
+
+void ambapp_prop_init(struct prop_std *buf, const char *prop_str);
+int ambapp_prop_int(struct prop_std *buf, const char *prop_str, int value);
+int ambapp_prop_str(struct prop_std *buf, const char *prop_str, const char *val_str);
+int ambapp_prop_uintarray(struct prop_std *buf, const char *prop_str,
+			  unsigned int *array, int len);
+int ambapp_prop_empty(struct prop_std *buf, const char *prop_str);
+
+int canoc_ahb_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	int val;
+	char *str;
+	const char *strval;
+	const char *compatibles[] =  {"aeroflexgaisler,can_oc", "nxp,sja1000", NULL};
+	int i, len;
+	static char *prop_str_nxpfreq = NULL;
+
+	switch (index) {
+	case 0:
+		ambapp_prop_init(buf, prop_str_compatible);
+		str = (char *)&buf->v.data[0];
+		buf->length = 0;
+		for(i = 0; compatibles[i]; i++) {
+			strval = compatibles[i];
+			len = 1 + _strnlen(strval, 64);
+			_strcpy(str, (char *)strval);
+			str += len;
+			buf->length += len;
+		}
+		return sizeof(struct prop_data) + buf->length;
+
+	case 1:
+		PROM_STRDUP_ONCE(prop_str_nxpfreq, "nxp,external-clock-frequency");
+		val = ambapp_dev_freq((ambapp_dev *)core->ahbslv);
+		return ambapp_prop_int(buf, prop_str_nxpfreq, val);
+
+	default:
+		return 0;
+	}
+}
+
+
+int i2c_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	int val;
+	static char *prop_str_regshift = NULL;
+
+	switch (index) {
+	case 0:
+		return ambapp_prop_str(buf, prop_str_compatible,
+				       "aeroflexgaisler,i2cmst");
+	case 1:
+		/* Set "clock-frequency" */
+		val = ambapp_dev_freq((ambapp_dev *)core->apbslv);
+		return ambapp_prop_int(buf, prop_str_clockfrequency, val);
+	case 2:
+		/* Set "reg-shift" */
+		PROM_STRDUP_ONCE(prop_str_regshift, "reg-shift");
+		return ambapp_prop_int(buf, prop_str_regshift, 2);
+	default:
+		return 0;
+	}
+}
+
+int spi_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	int val;
+
+	switch (index) {
+	case 0:
+		return ambapp_prop_str(buf, prop_str_compatible,
+				       "aeroflexgaisler,spictrl");
+	case 1:
+		/* Set "clock-frequency" */
+		val = ambapp_dev_freq((ambapp_dev *)core->apbslv);
+		return ambapp_prop_int(buf, prop_str_clockfrequency, val);
+	default:
+		return 0;
+	}
+}
+
+#define GRGPIO_IMAP_MASK	0x1f1f1f1f
+
+#define GRGPIO_IRQMODE_STRAIGHT	1
+#define GRGPIO_IRQMODE_SINGLE	2
+#define GRGPIO_IRQMODE_MAP	3
+
+#define GRGPIO_MAX_PINS		32
+#define GRGPIO_IRQMAP_NOIRQ	0xffffffff
+
+struct grgpio_fixup_priv {
+	unsigned int base;
+	unsigned int nbits;
+	unsigned int irqmap[GRGPIO_MAX_PINS];
+};
+
+int grgpio_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	struct grgpio_fixup_priv *gp =
+		(struct grgpio_fixup_priv *)core->fixup_priv;
+	static char *prop_str_base;
+	static char *prop_str_nbits;
+	static char *prop_str_irqmap;
+	static char *prop_str_gpiocontroller;
+	static char *prop_str_ngpiocells;
+
+	switch (index) {
+	case 0:
+		PROM_STRDUP_ONCE(prop_str_base, "base");
+		return ambapp_prop_int(buf, prop_str_base, gp->base);
+	case 1:
+		PROM_STRDUP_ONCE(prop_str_nbits, "nbits");
+		return ambapp_prop_int(buf, prop_str_nbits, gp->nbits);
+	case 2:
+		PROM_STRDUP_ONCE(prop_str_irqmap, "irqmap");
+		return ambapp_prop_uintarray(buf, prop_str_irqmap,
+					     gp->irqmap, gp->nbits);
+	case 3:
+		PROM_STRDUP_ONCE(prop_str_gpiocontroller, "gpiocontroller");
+		return ambapp_prop_empty(buf, prop_str_gpiocontroller);
+	case 4:
+		PROM_STRDUP_ONCE(prop_str_ngpiocells, "#gpio-cells");
+		return ambapp_prop_int(buf, prop_str_ngpiocells, 2);
+	default:
+		return 0;
+	}
+}
+
+/* Probe grgpio for nbits and interrupts */
+int grgpio_fixup_core(ambapp_core *core)
+{
+	int i, j;
+	int irqmode;
+	struct grgpio_regs {
+		unsigned int data;		/* 0x00 */
+		unsigned int output;		/* 0x04 */
+		unsigned int dir;		/* 0x08 */
+		unsigned int imask;		/* 0x0c */
+		unsigned int ipolarity;		/* 0x10  */
+		unsigned int iedge;		/* 0x14 */
+		unsigned int bypass;		/* 0x18 */
+		unsigned int __reserved;	/* 0x1c */
+		unsigned int imap[8];		/* 0x20-0x3c */
+	} *regs;
+
+	unsigned int imask;
+	unsigned int irqenabled;
+	unsigned int direction;
+	unsigned int output;
+	unsigned int lines;
+
+	/* Get registers */
+	struct grgpio_fixup_priv *gp = startup_malloc(sizeof(*gp));
+	ambapp_apbdev *dev = core->apbslv;
+
+	/* For keeping track of the running total of number of gpios over all
+	 * cores
+	 */
+	static unsigned int totalnbits = 0;
+
+	regs = (struct grgpio_regs *)dev->address;
+
+
+	/* Probe nbits, i.e. the number of lines */
+	direction = BYPASS_LOAD_PA(&regs->dir);
+	output = BYPASS_LOAD_PA(&regs->output);
+	BYPASS_STORE_PA(&regs->output, ~direction);
+	lines = direction | BYPASS_LOAD_PA(&regs->output);
+	BYPASS_STORE_PA(&regs->output, output);
+
+	/* Calculate nbits. Count the number of set bits in lines by
+	 * interativerly clearing the least significant bit */
+	for (gp->nbits = 0; lines; gp->nbits++)
+	{
+		lines &= lines - 1;
+	}
+	gp->base = totalnbits;
+	totalnbits += gp->nbits;
+
+	/* Probe which pins have irq enabled */
+	imask = BYPASS_LOAD_PA(&regs->imask);
+	BYPASS_STORE_PA(&regs->imask, 0xffffffff);
+	irqenabled = BYPASS_LOAD_PA(&regs->imask);
+	BYPASS_STORE_PA(&regs->imask, imask);
+
+	irqmode = 0;
+#ifdef CONFIG_GRGPIO_SINGLEIRQ
+	/* irqgen == 1 cannot be probed without going outside the grgpio
+	 * core. This has to be configured from the outside instead */
+	irqmode = GRGPIO_IRQMODE_SINGLE;
+#else
+
+
+	/* Probe if imap register address wrap around to general registers */
+	if (irqenabled) {
+		unsigned int oldval;
+
+		imask = BYPASS_LOAD_PA(&regs->imask);
+		oldval = BYPASS_LOAD_PA(&regs->imap[3]);
+		BYPASS_STORE_PA(&regs->imap[3], ~imask);
+		if (BYPASS_LOAD_PA(&regs->imask) != imask) {
+			irqmode = GRGPIO_IRQMODE_STRAIGHT;
+			BYPASS_STORE_PA(&regs->imask, imask);
+		}
+		BYPASS_STORE_PA(&regs->imap[3], oldval);
+	}
+
+	/* Probe if we have writable imap registers */
+	for (j = 0; j < 8 && !irqmode; j++) {
+		if ((0xf << (j*4)) & irqenabled) {
+			unsigned int read, readback;
+			void *address = &regs->imap[j];
+
+			read = BYPASS_LOAD_PA(address);
+			BYPASS_STORE_PA(address, ~read & GRGPIO_IMAP_MASK);
+			readback = BYPASS_LOAD_PA(address);
+			if(read != readback) {
+				irqmode = GRGPIO_IRQMODE_MAP;
+				BYPASS_STORE_PA(address, read);
+				break;
+			}
+		}
+	}
+	if(!irqmode) {
+		irqmode = GRGPIO_IRQMODE_STRAIGHT;
+	}
+#endif
+
+	/* Debug printout */
+	char *irqmodestr;
+	switch (irqmode) {
+	case GRGPIO_IRQMODE_STRAIGHT:
+		irqmodestr = "straight";
+		break;
+	case GRGPIO_IRQMODE_SINGLE:
+		irqmodestr = "single";
+		break;
+	case GRGPIO_IRQMODE_MAP:
+		irqmodestr = "map";
+		break;
+	default:
+		irqmodestr = "ERROR: undefined";
+		break;
+	}
+
+	/* Fill in irq arrays */
+	int pirq = dev->irq[0];
+	int irqindex = 0;
+	for (i = 0; i < gp->nbits; i++) {
+		if (irqenabled & (1 << i)) {
+			if (irqmode == GRGPIO_IRQMODE_SINGLE) {
+				/* All irq enabled lines have index 0 */
+				gp->irqmap[i] = 0;
+			} else {
+				int off;
+				if (irqmode == GRGPIO_IRQMODE_MAP) {
+					/* All have index 0 */
+					unsigned int map;
+					int j = i / 4;
+					map = BYPASS_LOAD_PA(&regs->imap[j]);
+					off = (map >> (8*(3-i))) & 0x1f;
+				} else {
+					off = i;
+				}
+				dev->irq[irqindex] = pirq + off;
+				gp->irqmap[i] = irqindex;
+				irqindex++;
+			}
+		} else {
+			gp->irqmap[i] = GRGPIO_IRQMAP_NOIRQ;
+		}
+	}
+
+	for (i = 0; i < 32 && dev->irq[i] != 0; i++) {
+		/* Check if dev->irq[i] overlaps with an eirq */
+		for (j = 0; eirqs[j]; j++)
+			if (dev->irq[i] == eirqs[j])
+				break;
+		if (!eirqs[j])
+			continue;
+
+		/* Overlap with eirq - remove from mapping... */
+		for (j = 0; j < GRGPIO_MAX_PINS; j++) {
+			if (gp->irqmap[j] == i) {
+				gp->irqmap[j] = GRGPIO_IRQMAP_NOIRQ;
+			} else if(gp->irqmap[j] > i && gp->irqmap[j] < 32) {
+				gp->irqmap[j]--; /* due to next step */
+			}
+		}
+
+		/* ... and from interrupts */
+		for (j = i+1; j < 32 && dev->irq[j] != 0; j++)
+			dev->irq[j-1] = dev->irq[j];
+		dev->irq[j-1] = 0;
+	}
+
+	core->fixup_priv = gp;
+
+	return 0;
+}
+
 void gptimer_init(ambapp_core *core, int level)
 {
 	/* Only process the first GPTIMER core */
@@ -114,14 +479,16 @@ void gptimer_init(ambapp_core *core, int level)
 #endif
 }
 
-static int apbuart_initied = CONFIG_UART_INDEX + 1;
+static int apbuart_initied = 0;
 
 /* Find first APBUART for console */
-void apbuart_init(ambapp_core *core, int level)
+int apbuart_apb_fixup(ambapp_dev *dev)
 {
-	if (apbuart_initied != 0 && (--apbuart_initied == 0) && core->apbslv) {
-		apbuart = (struct apbuart_regs *)core->apbslv->address;
+	if (!apbuart_initied++) {
+		ambapp_apbdev *apbdev = (ambapp_apbdev *)dev;
+		apbuart = (struct apbuart_regs *)apbdev->address;
 	}
+	return 0;
 }
 
 /* Turn off GRETH RX/TM DMA operations. This should not have to be done here,
@@ -146,36 +513,138 @@ void greth_init(ambapp_core *core, int level)
 	}
 }
 
+#define GR712RC_SYSTEMID 0x7120e70
+
+void canmux_init(ambapp_core *core, int level)
+{
+	struct canmux_regs {
+		uint32_t ctrl;
+	} *regs;
+
+	unsigned long systemid =
+		BYPASS_LOAD_PA(CONFIG_AMBA_IO_AREA | AMBA_CONF_AREA | 0xff0);
+
+	if (level == 0 && core->apbslv && systemid == GR712RC_SYSTEMID) {
+		regs = (struct canmux_regs *)core->apbslv->address;
+
+		/* Use both CAN_OC cores */
+		BYPASS_STORE_PA(&regs->ctrl, 3);
+	}
+}
+
 struct amba_driver {
 	unsigned short vendor;
 	unsigned short device;
 	unsigned char ver_min;
 	unsigned char ver_max;
-	unsigned char drv_type; /* 0x1=MST, 0x2=AHBSLV, 0x4=APBSLV, 0x8=CORE */
 	int (*fixup_dev)(ambapp_dev *dev);
 	int (*fixup_core)(ambapp_core *core);
+	int (*fixup_prop)(ambapp_core *core, struct prop_std *buf, int index);
 	void (*init)(ambapp_core *core, int level);
+	unsigned int (*fixup_dev_count)(struct ambapp_pnp_axb *axb, void *bus,
+					int type);
+	int (*fixup_virtual_dev)(ambapp_dev *dev, void *axb, void *bus,
+				 void **pdevp);
 	ambapp_core *cores;
 };
 
 /* Driver for a special version */
-#define AMBAPP_DRV_VER(ven, dev, vmin, vmax, type, fixdev, fixcore, init) \
-		{ven, dev, vmin, vmax, type, fixdev, fixcore, init, NULL}
+#define AMBAPP_DRV_VER(ven, dev, vmin, vmax, fixdev, fixcore, fixprop, init) \
+	{ven, dev, vmin, vmax, fixdev, fixcore, fixprop, init, NULL, NULL, NULL}
 /* Driver for all versions */
-#define AMBAPP_DRV(ven, dev, type, fixdev, fixcore, init) \
-		{ven, dev, 0, 0xff, type, fixdev, fixcore, init, NULL}
+#define AMBAPP_DRV(ven, dev, fixdev, fixcore, fixprop, init) \
+	{ven, dev, 0, 0xff, fixdev, fixcore, fixprop, init, NULL, NULL, NULL}
+/* Driver for all versions with extra fixup functions */
+#define AMBAPP_DRV_EXTRA_FIXUP(ven, dev, fixdev, fixcore, fixprop, init, \
+			       fixdevcnt, fixvirtdev)                   \
+	{ven, dev, 0, 0xff, fixdev, fixcore, fixprop, init,             \
+			fixdevcnt, fixvirtdev, NULL}
 
 struct amba_driver amba_drivers[] = {
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_IRQMP, 8, NULL, NULL, irqmp_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GPTIMER, 8, gptimer_apb_fixup, NULL, gptimer_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON4, 8, NULL, NULL, leon_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON3, 8, NULL, NULL, leon_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON3FT, 8, NULL, NULL, leon_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_APBUART, 8, NULL, NULL, apbuart_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_ETHMAC, 8, NULL, NULL, greth_init),
+ /*					   dev, core, prop, init */
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_IRQMP, irqmp_dev, NULL, NULL, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GPTIMER, gptimer_apb_fixup, NULL, NULL, gptimer_init),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON4, NULL, NULL, NULL, leon_init),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON3, NULL, NULL, NULL, leon_init),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON3FT, NULL, NULL, NULL, leon_init),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_APBUART, apbuart_apb_fixup, NULL, NULL, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_ETHMAC, NULL, NULL, NULL, greth_init),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_I2CMST, NULL, NULL, i2c_fixup_prop, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_SPICTRL, NULL, NULL, spi_fixup_prop, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GPIO, NULL, grgpio_fixup_core, grgpio_fixup_prop, NULL),
+ AMBAPP_DRV_EXTRA_FIXUP(VENDOR_GAISLER, GAISLER_CANAHB, NULL, NULL,     \
+			canoc_ahb_fixup_prop, NULL,                     \
+			canoc_ahb_fixup_dev_count, canoc_ahb_fixup_virtual),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_CANMUX, NULL, NULL, NULL, canmux_init),
 };
 #define AMBAPP_DRV_NUM (sizeof(amba_drivers)/sizeof(struct amba_driver))
 
+/* Fixup the count of AMBA PnP devices
+ *
+ * Note: To remove the device, instead return 1 in ambapp_dixup_dev. This is to
+ * add to the number of devices in conjunction with ambapp_fixup_virtual_dev.
+ *
+ * axb   - struct ambapp_pnp_ahb * or struct ambapp_pnp_apb * for device
+ * bus   - struct ahb_bus * or struct apb_bus * for device
+ * type  - Type of the device
+ *
+ * Return values:
+ *  The number of extra devices that will created.
+ */
+unsigned int ambapp_fixup_dev_count(struct ambapp_pnp_axb *axb, void *bus, int type)
+{
+	struct amba_driver *drv;
+	int i;
+
+	ambapp_dev dev_;
+	ambapp_dev *dev = &dev_;
+	dev->vendor = ambapp_pnp_vendor(axb->id);
+	dev->device = ambapp_pnp_device(axb->id);
+	dev->ver = ambapp_pnp_ver(axb->id);
+
+	for (i = 0, drv = amba_drivers; i < AMBAPP_DRV_NUM; i++, drv++) {
+		if ((drv->device == dev->device) &&
+		    (drv->fixup_dev_count != NULL) &&
+		    (drv->vendor == dev->vendor) &&
+		    ((dev->ver >= drv->ver_min) &&
+		     (dev->ver <= drv->ver_max))) {
+			return drv->fixup_dev_count(axb, bus, type);
+		}
+	}
+
+	return 0;
+}
+
+/* Fixup virtual AMBA PnP device information
+ *
+ * dev   - device struct for actual device
+ * axb   - struct ambapp_pnp_ahb * or struct ambapp_pnp_apb * for actual device
+ * bus   - struct ahb_bus * or struct apb_bus * for actual device
+ * pdevp - pointer to struct ambapp_ahbdev *pdev in ambapp_scan_bus_ahb, or
+ *         pointer to struct ambapp_apbdev *pdev in ambapp_scan_bus_apb
+ * Return values:
+ *  0 = Proceed as usual
+ *  1 = Skip original device but do not touch variables
+ */
+int ambapp_fixup_virtual_dev(ambapp_dev *dev, void *axb, void *bus,
+			     void **pdevp)
+{
+	struct amba_driver *drv;
+	int i;
+
+	for (i = 0, drv = amba_drivers; i < AMBAPP_DRV_NUM; i++, drv++) {
+		if ((drv->device == dev->device) &&
+		    (drv->fixup_virtual_dev != NULL) &&
+		    (drv->vendor == dev->vendor) &&
+		    ((dev->ver >= drv->ver_min) &&
+		     (dev->ver <= drv->ver_max))) {
+			return drv->fixup_virtual_dev(dev, axb, bus, pdevp);
+		}
+	}
+
+	return 0;
+}
+
 /* Fixup AMBA PnP device information
  *
  * Return values:
@@ -235,6 +704,36 @@ int ambapp_fixup_core(ambapp_core *core)
 	return 0;
 }
 
+/* Add properties for an ambapp_core
+ *
+ * core  - the core in question
+ * buf   - property to fill in when adding a new property
+ * index - for each core, this function should be called with an incrementing
+ *         index (starting at 0) until 0 is returned
+ *
+ * Return values:
+ *  0 = No more properties to be added
+ * >0 = The size of the property struct to be added
+ */
+int ambapp_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	struct amba_driver *drv;
+	int i;
+	ambapp_dev *dev = ambapp_get_first_dev(core);
+
+	for (i = 0, drv = amba_drivers; i < AMBAPP_DRV_NUM; i++, drv++) {
+		if ((drv->device == dev->device) &&
+		    (drv->fixup_prop != NULL) &&
+		    (drv->vendor == dev->vendor) &&
+		    ((dev->ver >= drv->ver_min) &&
+		     (dev->ver <= drv->ver_max))) {
+			return drv->fixup_prop(core, buf, index);
+		}
+	}
+
+	return 0;
+}
+
 int ambapp_node_fixup(ambapp_core *core, struct node *node)
 {
 	ambapp_dev *dev = ambapp_get_first_dev(core);
@@ -608,8 +1107,10 @@ static void ambapp_scan_bus_ahb(
 	)
 {
 	struct ambapp_pnp_ahb *ahbs[64], ahb_buf, **pahb;
+	struct ambapp_pnp_axb *axb;
 	ambapp_ahbdev *pdev;
-	int i, size, max_entries;
+	int i, size, max_entries, vfix;
+	unsigned int delta_count, pnp_cnt;
 	unsigned long ioarea;
 	struct ambapp_bridge_driver *busdrv;
 
@@ -622,48 +1123,79 @@ static void ambapp_scan_bus_ahb(
 	/* AHB MASTERS */
 	ioarea = bus->ioarea | AMBA_CONF_AREA;
 	size = sizeof(ambapp_ahbdev);
-	bus->mst_cnt = ambapp_scan_bus(ioarea, max_entries,
-				sizeof(struct ambapp_pnp_ahb), (void **)ahbs,
-				&size);
+	pnp_cnt = ambapp_scan_bus(ioarea, max_entries,
+				  sizeof(struct ambapp_pnp_ahb), (void **)ahbs,
+				  &size);
+
+	delta_count = 0;
+	for (i = 0; i < pnp_cnt; i++) {
+		_memcpy4_pa(&ahb_buf, ahbs[i], sizeof(struct ambapp_pnp_ahb));
+		axb = (struct ambapp_pnp_axb *)&ahb_buf;
+		delta_count += ambapp_fixup_dev_count(axb, bus, DEV_AHB_MST);
+	}
+	bus->mst_cnt = pnp_cnt + delta_count;
+	size += delta_count * sizeof(ambapp_ahbdev);
 	bus->msts = (ambapp_ahbdev *)startup_malloc(size);
 
 	for (i = 0, pahb = &ahbs[0], pdev = &bus->msts[0];
-		i < bus->mst_cnt; i++, pahb++, pdev++) {
+	     i < pnp_cnt; i++, pahb++, pdev++) {
 		_memcpy4_pa(&ahb_buf, *pahb, sizeof(struct ambapp_pnp_ahb));
 
 		/* Build a AHB Master Device */
 		pdev->dev_type = DEV_AHB_MST;
 		ambapp_dev_init_ahb(bus, &ahb_buf, pdev);
 
+		vfix = ambapp_fixup_virtual_dev((ambapp_dev *)pdev,
+						(void*)&ahb_buf, (void*)bus,
+						(void**)&pdev);
+		if (vfix == 1)
+			/* skip the rest of the loop */
+			continue;
+
 		if (ambapp_fixup_dev((ambapp_dev *)pdev) == 1) {
 			/* skip device */
 			bus->mst_cnt--;
 			pdev--;
-			i--;
 			continue;
 		}
 	}
 
+	/* AHB SLAVES */
 	ioarea |= AMBA_AHBSLV_AREA;
 	size = sizeof(ambapp_ahbdev);
-	bus->slv_cnt = ambapp_scan_bus(ioarea, max_entries,
-				sizeof(struct ambapp_pnp_ahb), (void **)ahbs,
-				&size);
+	pnp_cnt = ambapp_scan_bus(ioarea, max_entries,
+				  sizeof(struct ambapp_pnp_ahb), (void **)ahbs,
+				  &size);
+
+	delta_count = 0;
+	for (i = 0; i < pnp_cnt; i++) {
+		_memcpy4_pa(&ahb_buf, ahbs[i], sizeof(struct ambapp_pnp_ahb));
+		axb = (struct ambapp_pnp_axb *)&ahb_buf;
+		delta_count += ambapp_fixup_dev_count(axb, bus, DEV_AHB_SLV);
+	}
+	bus->slv_cnt = pnp_cnt + delta_count;
+	size += delta_count * sizeof(ambapp_ahbdev);
 	bus->slvs = (ambapp_ahbdev *)startup_malloc(size);
 
 	for (i = 0, pahb = &ahbs[0], pdev = &bus->slvs[0];
-		i < bus->slv_cnt; i++, pahb++, pdev++) {
+	     i < pnp_cnt; i++, pahb++, pdev++) {
 		_memcpy4_pa(&ahb_buf, *pahb, sizeof(struct ambapp_pnp_ahb));
 
 		/* Build a AHB Slave Device */
 		pdev->dev_type = DEV_AHB_SLV;
 		ambapp_dev_init_ahb(bus, &ahb_buf, pdev);
 
+		vfix = ambapp_fixup_virtual_dev((ambapp_dev *)pdev,
+						(void*)&ahb_buf, (void*)bus,
+						(void**)&pdev);
+		if (vfix == 1)
+			/* skip the rest of the loop */
+			continue;
+
 		if (ambapp_fixup_dev((ambapp_dev *)pdev) == 1) {
 			/* skip device */
 			bus->slv_cnt--;
 			pdev--;
-			i--;
 			continue;
 		}
 
@@ -679,31 +1211,48 @@ static void ambapp_scan_bus_ahb(
 static void ambapp_scan_bus_apb(struct ambapp_bus *ainfo, struct apb_bus *bus)
 {
 	struct ambapp_pnp_apb *apbs[AMBA_APB_SLAVES], apb_buf, **papb;
+	struct ambapp_pnp_axb *axb;
 	ambapp_apbdev *pdev;
-	int i, size;
+	int i, size, vfix;
+	unsigned int delta_count, pnp_cnt;
 
-	/* AHB MASTERS */
+	/* APB SLAVES */
 	size = sizeof(ambapp_apbdev);
-	bus->slv_cnt = ambapp_scan_bus(bus->ioarea | AMBA_CONF_AREA,
-					AMBA_APB_SLAVES,
-					sizeof(struct ambapp_pnp_apb),
-					(void **)apbs,
-					&size);
+	pnp_cnt = ambapp_scan_bus(bus->ioarea | AMBA_CONF_AREA,
+				  AMBA_APB_SLAVES,
+				  sizeof(struct ambapp_pnp_apb),
+				  (void **)apbs,
+				  &size);
+
+	delta_count = 0;
+	for (i = 0; i < pnp_cnt; i++) {
+		_memcpy4_pa(&apb_buf, apbs[i], sizeof(struct ambapp_pnp_apb));
+		axb = (struct ambapp_pnp_axb *)&apb_buf;
+		delta_count += ambapp_fixup_dev_count(axb, bus, DEV_APB_SLV);
+	}
+	bus->slv_cnt = pnp_cnt + delta_count;
+	size += delta_count * sizeof(ambapp_apbdev);
 	bus->slvs = (ambapp_apbdev *)startup_malloc(size);
 
 	for (i = 0, papb = &apbs[0], pdev = &bus->slvs[0];
-		i < bus->slv_cnt; i++, papb++, pdev++) {
+		i < pnp_cnt; i++, papb++, pdev++) {
 		_memcpy4_pa(&apb_buf, *papb, sizeof(struct ambapp_pnp_apb));
 
-		/* Build a AHB Slave Device */
+		/* Build a APB Slave Device */
 		pdev->dev_type = DEV_APB_SLV;
 		ambapp_dev_init_apb(bus, &apb_buf, pdev);
 
+		vfix = ambapp_fixup_virtual_dev((ambapp_dev *)pdev,
+						(void*)&apb_buf, (void*)bus,
+						(void**)&pdev);
+		if (vfix == 1)
+			/* skip the rest of the loop */
+			continue;
+
 		if (ambapp_fixup_dev((ambapp_dev *)pdev) == 1) {
 			/* skip device */
 			bus->slv_cnt--;
 			pdev--;
-			i--;
 			continue;
 		}
 
@@ -1114,6 +1663,56 @@ void writehex(char *n, int v)
 	*n = v > 9 ? v - 10 + 'a' : v + '0';
 }
 
+void ambapp_prop_init(struct prop_std *buf, const char *prop_str)
+{
+	struct prop_data *dp = (struct prop_data *)buf;
+
+	dp->name = (char *)prop_str;
+	dp->next = NULL;
+	dp->options = PO_DATA | PO_NEXT;
+}
+
+int ambapp_prop_int(struct prop_std *buf, const char *prop_str, int value)
+{
+	struct prop_data *dp = (struct prop_data *)buf;
+
+	ambapp_prop_init(buf, prop_str);
+	dp->data[0] = value;
+	dp->length = sizeof(value);
+
+	return sizeof(struct prop_data) + dp->length;
+}
+
+int ambapp_prop_str(struct prop_std *buf, const char *prop_str, const char *val_str)
+{
+	char *str = (char *)&buf->v.data[0];
+
+	if (!val_str)
+		return 0;
+	_strcpy(str, (char *)val_str);
+	buf->length = 1 + _strnlen(str, 64);
+	ambapp_prop_init(buf, prop_str);
+
+	return sizeof(struct prop_data) + buf->length;
+}
+
+int ambapp_prop_uintarray(struct prop_std *buf, const char *prop_str,
+			  unsigned int *array, int len)
+{
+	struct prop_data *dp = (struct prop_data *)buf;
+
+	ambapp_prop_init(buf, prop_str);
+	dp->length = len * sizeof(*array);
+	_memcpy((char *)dp->data, (char *)array, dp->length);
+
+	return sizeof(struct prop_data) + dp->length;
+}
+
+int ambapp_prop_empty(struct prop_std *buf, const char *prop_str)
+{
+	return ambapp_prop_uintarray(buf, prop_str, NULL, 0);
+}
+
 /* Create Name Property */
 int ambapp_prop_name(ambapp_core *core, struct prop_std *buf)
 {
@@ -1144,76 +1743,43 @@ int ambapp_prop_name(ambapp_core *core, struct prop_std *buf)
 		buf->length = 7;
 	}
 
-	buf->name = prop_str_name;
-	buf->next = NULL;
-	buf->options = PO_DATA | PO_NEXT;
+	ambapp_prop_init(buf, prop_str_name);
 
 	return sizeof(struct prop_data) + buf->length;
 }
 
 int ambapp_prop_vendor(ambapp_core *core, struct prop_std *buf)
 {
-	struct prop_data *dp = (struct prop_data *)buf;
 	ambapp_dev *dev = ambapp_get_first_dev(core);
 
-	dp->name = prop_str_vendor;
-	dp->next = NULL;
-	dp->length = 4;
-	dp->options = PO_DATA | PO_NEXT;
-	dp->data[0] = dev->vendor;
-
-	return sizeof(struct prop_data) + sizeof(int);
+	return ambapp_prop_int(buf, prop_str_vendor, dev->vendor);
 }
 
 int ambapp_prop_device(ambapp_core *core, struct prop_std *buf)
 {
-	struct prop_data *dp = (struct prop_data *)buf;
 	ambapp_dev *dev = ambapp_get_first_dev(core);
 
-	dp->name = prop_str_device;
-	dp->next = NULL;
-	dp->length = 4;
-	dp->options = PO_DATA | PO_NEXT;
-	dp->data[0] = dev->device;
-
-	return sizeof(struct prop_data) + sizeof(int);
+	return ambapp_prop_int(buf, prop_str_device, dev->device);
 }
 
 int ambapp_prop_version(ambapp_core *core, struct prop_std *buf)
 {
-	struct prop_data *dp = (struct prop_data *)buf;
 	ambapp_dev *dev = ambapp_get_first_dev(core);
 
-	dp->name = prop_str_version;
-	dp->next = NULL;
-	dp->length = 4;
-	dp->options = PO_DATA | PO_NEXT;
-	dp->data[0] = dev->ver;
-
-	return sizeof(struct prop_data) + sizeof(int);
+	return ambapp_prop_int(buf, prop_str_version, dev->ver);
 }
 
 int ambapp_prop_devmsk(ambapp_core *core, struct prop_std *buf)
 {
-	struct prop_data *dp = (struct prop_data *)buf;
-	unsigned int devmsk;
+	unsigned int devmsk = 0;
 
-	dp->name = prop_str_devmsk;
-	dp->next = NULL;
-	dp->length = 4;
-	dp->options = PO_DATA | PO_NEXT;
-
-	devmsk = 0;
 	if (core->apbslv)
 		devmsk |= 0x1;
 	if (core->ahbslv)
 		devmsk |= 0x2;
 	if (core->ahbmst)
 		devmsk |= 0x4;
-
-	dp->data[0] = devmsk;
-
-	return sizeof(struct prop_data) + sizeof(int);
+	return ambapp_prop_int(buf, prop_str_devmsk, (int)devmsk);
 }
 
 int ambapp_prop_interrupts_add(ambapp_dev *dev, int *buf, int index)
@@ -1241,9 +1807,7 @@ int ambapp_prop_interrupts(ambapp_core *core, struct prop_std *buf)
 	if (i == 0)
 		return 0; /* no such property */
 
-	dp->name = prop_str_interrupts;
-	dp->next = NULL;
-	dp->options = PO_DATA | PO_NEXT;
+	ambapp_prop_init(buf, prop_str_interrupts);
 	dp->length = i << 2;
 
 	return sizeof(struct prop_data) + dp->length;
@@ -1271,9 +1835,8 @@ int ambapp_prop_reg(ambapp_core *core, struct prop_std *buf)
 	if (i == 0)
 		return 0; /* no reg property */
 
-	dp->name = prop_str_reg;
-	dp->next = NULL;
-	dp->options = PO_DATA | PO_NEXT;
+
+	ambapp_prop_init(buf, prop_str_reg);
 	dp->length = i << 2;
 
 	return sizeof(struct prop_data) + dp->length;
@@ -1299,9 +1862,7 @@ int ambapp_prop_userdef(ambapp_core *core, struct prop_std *buf)
 	if (i == 0)
 		return 0; /* no reg property */
 
-	dp->name = prop_str_userdef;
-	dp->next = NULL;
-	dp->options = PO_DATA | PO_NEXT;
+	ambapp_prop_init(buf, prop_str_userdef);
 	dp->length = i << 2;
 
 	return sizeof(struct prop_data) + dp->length;
@@ -1312,9 +1873,7 @@ int ambapp_prop_freq(ambapp_core *core, struct prop_std *buf)
 	int i;
 	struct prop_data *dp = (struct prop_data *)buf;
 
-	dp->name = prop_str_freq;
-	dp->next = NULL;
-	dp->options = PO_DATA | PO_NEXT;
+	ambapp_prop_init(buf, prop_str_freq);
 	i = 0;
 	if (core->apbslv)
 		dp->data[i++] = ambapp_dev_freq((ambapp_dev *)core->apbslv);
@@ -1329,15 +1888,7 @@ int ambapp_prop_freq(ambapp_core *core, struct prop_std *buf)
 
 int ambapp_prop_index(ambapp_core *core, struct prop_std *buf)
 {
-	struct prop_data *dp = (struct prop_data *)buf;
-
-	dp->name = prop_str_index;
-	dp->next = NULL;
-	dp->options = PO_DATA | PO_NEXT;
-	dp->data[0] = core->index;
-	dp->length = sizeof(int);
-
-	return sizeof(struct prop_data) + sizeof(int);
+	return ambapp_prop_int(buf, prop_str_index, core->index);
 }
 
 int ambapp_prop_ampopts(ambapp_core *core, struct prop_std *buf)
@@ -1351,15 +1902,8 @@ int ambapp_prop_ampopts(ambapp_core *core, struct prop_std *buf)
 	while (a->idx != -1) {
 		if (a->idx == core->index) {
 			/* Found Matching AMP String, create property */
-			struct prop_data *dp = (struct prop_data *)buf;
-
-			dp->name = prop_str_ampopts;
-			dp->next = NULL;
-			dp->options = PO_DATA | PO_NEXT;
-			dp->data[0] = a->val;
-			dp->length = sizeof(int);
-
-			return sizeof(struct prop_data) + sizeof(int);
+			return ambapp_prop_int(buf, prop_str_ampopts,
+					       a->val);
 		}
 		a++;
 	}
@@ -1372,25 +1916,11 @@ int ambapp_prop_ampopts(ambapp_core *core, struct prop_std *buf)
 /* Create Description Property... very optional :) */
 int ambapp_prop_description(ambapp_core *core, struct prop_std *buf)
 {
-	char *str;
-	const char *desc;
-	ambapp_dev *dev = ambapp_get_first_dev(core);
-
-	str = (char *)&buf->v.data[0];
-
 	/* Find Device description from AMBA PnP Data Base */
-	desc = ambapp_device_id2desc(dev->vendor, dev->device);
-	if (desc == NULL)
-		return 0;
+	ambapp_dev *dev = ambapp_get_first_dev(core);
+	const char *desc = ambapp_device_id2desc(dev->vendor, dev->device);
 
-	_strcpy(str, (char *)desc);
-	buf->length = 1 + _strnlen(desc, 64);
-
-	buf->name = prop_str_description;
-	buf->next = NULL;
-	buf->options = PO_DATA | PO_NEXT;
-
-	return sizeof(struct prop_data) + buf->length;
+	return ambapp_prop_str(buf, prop_str_description, desc);
 }
 
 int (*ambapp_prop_creators[])(ambapp_core *core, struct prop_std *buf) = {
@@ -1443,7 +1973,7 @@ struct node *ambapp_node_create(ambapp_core *core)
 {
 	struct node *n;
 	int i, size;
-	unsigned int prop_buf[32];
+	unsigned int prop_buf[64];
 	struct prop_std *prop, *buf = (struct prop_std *)prop_buf;
 
 	n = (struct node *)prom_malloc(sizeof(struct node));
@@ -1460,6 +1990,15 @@ struct node *ambapp_node_create(ambapp_core *core)
 		_memcpy((void *)prop, (void *)buf, size);
 		prop_add(n, (struct prop *)prop);
 	}
+	for (i = 0; ; i++) {
+		size = ambapp_fixup_prop(core, buf, i);
+		if (size == 0)
+			break;
+		prop = (struct prop_std *)prom_malloc(size);
+		_memcpy((void *)prop, (void *)buf, size);
+		prop_add(n, (struct prop *)prop);
+	}
+
 
 #ifdef HASXML
 	{
@@ -1485,6 +2024,8 @@ struct node *ambapp_node_create(ambapp_core *core)
 			} else { /* add node-tree */
 				struct mknod_def *nd = nod;
 				for (k = s->from; k < s->to; k++) {
+					if (nd[k].lev > 1)
+						continue;
 					cn = (struct node *)prom_calloc(sizeof(struct node));
 					struct node **np = &n->child;
 					while (*np)
